=== modified file 'NEWS'
--- NEWS	2009-01-29 16:46:44 +0000
+++ NEWS	2009-01-31 04:17:43 +0000
@@ -46,6 +46,10 @@
     * ``bzr init`` will now print a little less verbose output.
       (Marius Kruger)
 
+    * ``bzr log`` is now much faster in many use cases, particularly
+      at incrementally displaying results and filtering by a
+      revision range. (Ian Clatworthy)
+
     * ``bzr log --short`` and ``bzr log --line`` now show tags, if any,
       for each revision. The tags are shown comma-separated inside
       ``{}``. For short format, the tags appear at the end of line

=== modified file 'bzrlib/builtins.py'
--- bzrlib/builtins.py	2009-01-29 07:57:43 +0000
+++ bzrlib/builtins.py	2009-02-02 05:43:13 +0000
@@ -49,7 +49,7 @@
     )
 from bzrlib.branch import Branch
 from bzrlib.conflicts import ConflictList
-from bzrlib.revisionspec import RevisionSpec
+from bzrlib.revisionspec import RevisionSpec, RevisionInfo
 from bzrlib.smtp_connection import SMTPConnection
 from bzrlib.workingtree import WorkingTree
 """)
@@ -1972,16 +1972,23 @@
     elif len(revisionspec_list) == 1:
         rev1 = rev2 = revisionspec_list[0].in_history(branch)
     elif len(revisionspec_list) == 2:
-        if revisionspec_list[1].get_branch() != revisionspec_list[0
-                ].get_branch():
+        start_spec = revisionspec_list[0]
+        end_spec = revisionspec_list[1]
+        if end_spec.get_branch() != start_spec.get_branch():
             # b is taken from revision[0].get_branch(), and
             # show_log will use its revision_history. Having
             # different branches will lead to weird behaviors.
             raise errors.BzrCommandError(
                 "bzr %s doesn't accept two revisions in different"
                 " branches." % command_name)
-        rev1 = revisionspec_list[0].in_history(branch)
-        rev2 = revisionspec_list[1].in_history(branch)
+        rev1 = start_spec.in_history(branch)
+        # Avoid loading all of history when we know a missing
+        # end of range means the last revision ...
+        if end_spec.spec is None:
+            last_revno, last_revision_id = branch.last_revision_info()
+            rev2 = RevisionInfo(branch, last_revno, last_revision_id)
+        else:
+            rev2 = end_spec.in_history(branch)
     else:
         raise errors.BzrCommandError(
             'bzr %s --revision takes one or two values.' % command_name)

=== modified file 'bzrlib/log.py'
--- bzrlib/log.py	2009-01-29 12:24:29 +0000
+++ bzrlib/log.py	2009-02-02 08:28:54 +0000
@@ -52,6 +52,7 @@
 import codecs
 from cStringIO import StringIO
 from itertools import (
+    chain,
     izip,
     )
 import re
@@ -199,34 +200,31 @@
     """Worker function for show_log - see show_log."""
     if not isinstance(lf, LogFormatter):
         warn("not a LogFormatter instance: %r" % lf)
-
     if specific_fileid:
         trace.mutter('get log for file_id %r', specific_fileid)
+
+    # Consult the LogFormatter about what it needs and can handle
     levels_to_display = lf.get_levels()
     generate_merge_revisions = levels_to_display != 1
     allow_single_merge_revision = True
     if not getattr(lf, 'supports_merge_revisions', False):
         allow_single_merge_revision = getattr(lf,
             'supports_single_merge_revision', False)
-    view_revisions = calculate_view_revisions(branch, start_revision,
-                                              end_revision, direction,
-                                              specific_fileid,
-                                              generate_merge_revisions,
-                                              allow_single_merge_revision)
-    rev_tag_dict = {}
     generate_tags = getattr(lf, 'supports_tags', False)
-    if generate_tags:
-        if branch.supports_tags():
-            rev_tag_dict = branch.tags.get_reverse_tag_dict()
-
+    if generate_tags and branch.supports_tags():
+        rev_tag_dict = branch.tags.get_reverse_tag_dict()
+    else:
+        rev_tag_dict = {}
     generate_delta = verbose and getattr(lf, 'supports_delta', False)
     generate_diff = show_diff and getattr(lf, 'supports_diff', False)
 
-    # now we just print all the revisions
+    # Find and print the interesting revisions
     repo = branch.repository
     log_count = 0
-    revision_iterator = make_log_rev_iterator(branch, view_revisions,
-        generate_delta, search)
+    revision_iterator = _create_log_revision_iterator(branch,
+        start_revision, end_revision, direction, specific_fileid, search,
+        generate_merge_revisions, allow_single_merge_revision,
+        generate_delta, limited_output=limit > 0)
     for revs in revision_iterator:
         for (rev_id, revno, merge_depth), rev, delta in revs:
             # Note: 0 levels means show everything; merge_depth counts from 0
@@ -262,48 +260,275 @@
     return s.getvalue()
 
 
-def calculate_view_revisions(branch, start_revision, end_revision, direction,
-                             specific_fileid, generate_merge_revisions,
-                             allow_single_merge_revision):
-    if (    not generate_merge_revisions
-        and start_revision is end_revision is None
-        and direction == 'reverse'
-        and specific_fileid is None):
-        return _linear_view_revisions(branch)
-
-    mainline_revs, rev_nos, start_rev_id, end_rev_id = _get_mainline_revs(
-        branch, start_revision, end_revision)
-    if not mainline_revs:
+class _StartNotLinearAncestor(Exception):
+    """Raised when a start revision is not found walking left-hand history."""
+
+
+def _create_log_revision_iterator(branch, start_revision, end_revision,
+    direction, specific_fileid, search, generate_merge_revisions,
+    allow_single_merge_revision, generate_delta, limited_output=False):
+    """Create a revision iterator for log.
+
+    :param branch: The branch being logged.
+    :param start_revision: If not None, only show revisions >= start_revision
+    :param end_revision: If not None, only show revisions <= end_revision
+    :param direction: 'reverse' (default) is latest to earliest; 'forward' is
+        earliest to latest.
+    :param specific_fileid: If not None, list only the commits affecting the
+        specified file.
+    :param search: If not None, only show revisions with matching commit
+        messages.
+    :param generate_merge_revisions: If False, show only mainline revisions.
+    :param allow_single_merge_revision: If True, logging of a single
+        revision off the mainline is to be allowed
+    :param generate_delta: Whether to generate a delta for each revision.
+    :param limited_output: if True, the user only wants a limited result
+
+    :return: An iterator over lists of ((rev_id, revno, merge_depth), rev,
+        delta).
+    """
+    start_rev_id, end_rev_id = _get_revision_limits(branch, start_revision,
+        end_revision)
+
+    # Decide how file-ids are matched: delta-filtering vs per-file graph.
+    # Delta filtering allows revisions to be displayed incrementally
+    # though the total time is much slower for huge repositories: log -v
+    # is the *lower* performance bound. At least until the split
+    # inventory format arrives, per-file-graph needs to remain the
+    # default except in verbose mode. Delta filtering should give more
+    # accurate results (e.g. inclusion of FILE deletions) so arguably
+    # it should always be used in the future.
+    use_deltas_for_matching = specific_fileid and generate_delta
+    delayed_graph_generation = not specific_fileid and (
+            start_rev_id or end_rev_id or limited_output)
+    generate_merges = generate_merge_revisions or (specific_fileid and
+        not use_deltas_for_matching)
+    view_revisions = _calc_view_revisions(branch, start_rev_id, end_rev_id,
+        direction, generate_merges, allow_single_merge_revision,
+        delayed_graph_generation=delayed_graph_generation)
+    search_deltas_for_fileids = None
+    if use_deltas_for_matching:
+        search_deltas_for_fileids = set([specific_fileid])
+    elif specific_fileid:
+        if not isinstance(view_revisions, list):
+            view_revisions = list(view_revisions)
+        view_revisions = _filter_revisions_touching_file_id(branch,
+            specific_fileid, view_revisions,
+            include_merges=generate_merge_revisions)
+    return make_log_rev_iterator(branch, view_revisions, generate_delta,
+        search, file_ids=search_deltas_for_fileids, direction=direction)
+
+
+def _calc_view_revisions(branch, start_rev_id, end_rev_id, direction,
+    generate_merge_revisions, allow_single_merge_revision,
+    delayed_graph_generation=False):
+    """Calculate the revisions to view.
+
+    :return: An iterator of (revision_id, dotted_revno, merge_depth) tuples OR
+             a list of the same tuples.
+    """
+    br_revno, br_rev_id = branch.last_revision_info()
+    if br_revno == 0:
         return []
 
-    generate_single_revision = False
-    if ((not generate_merge_revisions)
-        and ((start_rev_id and (start_rev_id not in rev_nos))
-            or (end_rev_id and (end_rev_id not in rev_nos)))):
-        generate_single_revision = ((start_rev_id == end_rev_id)
-            and allow_single_merge_revision)
-        if not generate_single_revision:
-            raise errors.BzrCommandError('Selected log formatter only supports'
-                ' mainline revisions.')
-        generate_merge_revisions = generate_single_revision
-    include_merges = generate_merge_revisions or specific_fileid
-    view_revs_iter = get_view_revisions(mainline_revs, rev_nos, branch,
-                          direction, include_merges=include_merges)
-
+    # If a single revision is requested, check we can handle it
+    generate_single_revision = (end_rev_id and start_rev_id == end_rev_id and
+        (not generate_merge_revisions or not _has_merges(branch, end_rev_id)))
+    if generate_single_revision:
+        if end_rev_id == br_rev_id:
+            # It's the tip
+            return [(br_rev_id, br_revno, 0)]
+        else:
+            revno = branch.revision_id_to_dotted_revno(end_rev_id)
+            if len(revno) > 1 and not allow_single_merge_revision:
+                # It's a merge revision and the log formatter is
+                # completely brain dead. This "feature" of allowing
+                # log formatters incapable of displaying dotted revnos
+                # ought to be deprecated IMNSHO. IGC 20091022
+                raise errors.BzrCommandError('Selected log formatter only'
+                    ' supports mainline revisions.')
+            revno_str = '.'.join(str(n) for n in revno)
+            return [(end_rev_id, revno_str, 0)]
+
+    # If we only want to see linear revisions, we can iterate ...
+    if not generate_merge_revisions:
+        result = _linear_view_revisions(branch, start_rev_id, end_rev_id)
+        # If a start limit was given and it's not obviously an
+        # ancestor of the end limit, check it before outputting anything
+        if direction == 'forward' or (start_rev_id
+            and not _is_obvious_ancestor(branch, start_rev_id, end_rev_id)):
+            try:
+                result = list(result)
+            except _StartNotLinearAncestor:
+                raise errors.BzrCommandError('Start revision not found in'
+                    ' left-hand history of end revision.')
+        if direction == 'forward':
+            result = reversed(list(result))
+        return result
+
+    # On large trees, generating the merge graph can take 30-60 seconds
+    # so we delay doing it until a merge is detected, incrementally
+    # returning initial (non-merge) revisions while we can.
+    initial_revisions = []
+    if delayed_graph_generation:
+        try:
+            for rev_id, revno, depth in \
+                _linear_view_revisions(branch, start_rev_id, end_rev_id):
+                if _has_merges(branch, rev_id):
+                    end_rev_id = rev_id
+                    break
+                else:
+                    initial_revisions.append((rev_id, revno, depth))
+            else:
+                # No merged revisions found
+                if direction == 'reverse':
+                    return initial_revisions
+                elif direction == 'forward':
+                    return reversed(initial_revisions)
+                else:
+                    raise ValueError('invalid direction %r' % direction)
+        except _StartNotLinearAncestor:
+            # A merge was never detected so the lower revision limit can't
+            # be nested down somewhere
+            raise errors.BzrCommandError('Start revision not found in'
+                ' history of end revision.')
+
+    # A log including nested merges is required. If the direction is reverse,
+    # we rebase the initial merge depths so that the development line is
+    # shown naturally, i.e. just like it is for linear logging. We can easily
+    # make forward the exact opposite display, but showing the merge revisions
+    # indented at the end seems slightly nicer in that case.
+    view_revisions = chain(iter(initial_revisions),
+        _graph_view_revisions(branch, start_rev_id, end_rev_id,
+        rebase_initial_depths=direction == 'reverse'))
     if direction == 'reverse':
-        start_rev_id, end_rev_id = end_rev_id, start_rev_id
-    view_revisions = _filter_revision_range(list(view_revs_iter),
-                                            start_rev_id,
-                                            end_rev_id)
-    if view_revisions and generate_single_revision:
-        view_revisions = view_revisions[0:1]
+        return view_revisions
+    elif direction == 'forward':
+        # Forward means oldest first, adjusting for depth.
+        view_revisions = reverse_by_depth(list(view_revisions))
+        return _rebase_merge_depth(view_revisions)
+    else:
+        raise ValueError('invalid direction %r' % direction)
+
+
+def _has_merges(branch, rev_id):
+    """Does a revision have multiple parents or not?"""
+    parents = branch.repository.get_parent_map([rev_id]).get(rev_id, [])
+    return len(parents) > 1
+
+
+def _is_obvious_ancestor(branch, start_rev_id, end_rev_id):
+    """Is start_rev_id an obvious ancestor of end_rev_id?"""
+    if start_rev_id and end_rev_id:
+        start_dotted = branch.revision_id_to_dotted_revno(start_rev_id)
+        end_dotted = branch.revision_id_to_dotted_revno(end_rev_id)
+        if len(start_dotted) == 1 and len(end_dotted) == 1:
+            # both on mainline
+            return start_dotted[0] <= end_dotted[0]
+        elif (len(start_dotted) == 3 and len(end_dotted) == 3 and
+            start_dotted[0:1] == end_dotted[0:1]):
+            # both on same development line
+            return start_dotted[2] <= end_dotted[2]
+        else:
+            # not obvious
+            return False
+    return True
+
+
+def _linear_view_revisions(branch, start_rev_id, end_rev_id):
+    """Calculate a sequence of revisions to view, newest to oldest.
+
+    :param start_rev_id: the lower revision-id
+    :param end_rev_id: the upper revision-id
+    :return: An iterator of (revision_id, dotted_revno, merge_depth) tuples.
+    :raises _StartNotLinearAncestor: if a start_rev_id is specified but
+      is not found walking the left-hand history
+    """
+    br_revno, br_rev_id = branch.last_revision_info()
+    repo = branch.repository
+    if start_rev_id is None and end_rev_id is None:
+        cur_revno = br_revno
+        for revision_id in repo.iter_reverse_revision_history(br_rev_id):
+            yield revision_id, str(cur_revno), 0
+            cur_revno -= 1
+    else:
+        if end_rev_id is None:
+            end_rev_id = br_rev_id
+        found_start = start_rev_id is None
+        for revision_id in repo.iter_reverse_revision_history(end_rev_id):
+            revno = branch.revision_id_to_dotted_revno(revision_id)
+            revno_str = '.'.join(str(n) for n in revno)
+            if not found_start and revision_id == start_rev_id:
+                yield revision_id, revno_str, 0
+                found_start = True
+                break
+            else:
+                yield revision_id, revno_str, 0
+        else:
+            if not found_start:
+                raise _StartNotLinearAncestor()
+
+
+def _graph_view_revisions(branch, start_rev_id, end_rev_id,
+    rebase_initial_depths=True):
+    """Calculate revisions to view including merges, newest to oldest.
+
+    :param branch: the branch
+    :param start_rev_id: the lower revision-id
+    :param end_rev_id: the upper revision-id
+    :param rebase_initial_depth: should depths be rebased until a mainline
+      revision is found?
+    :return: An iterator of (revision_id, dotted_revno, merge_depth) tuples.
+    """
+    view_revisions = branch.iter_merge_sorted_revisions(
+        start_revision_id=end_rev_id, stop_revision_id=start_rev_id,
+        stop_rule="with-merges")
+    if not rebase_initial_depths:
+        for (rev_id, merge_depth, revno, end_of_merge
+             ) in view_revisions:
+            yield rev_id, '.'.join(map(str, revno)), merge_depth
+    else:
+        # We're following a development line starting at a merged revision.
+        # We need to adjust depths down by the initial depth until we find
+        # a depth less than it. Then we use that depth as the adjustment.
+        # If and when we reach the mainline, depth adjustment ends.
+        depth_adjustment = None
+        for (rev_id, merge_depth, revno, end_of_merge
+             ) in view_revisions:
+            if depth_adjustment is None:
+                depth_adjustment = merge_depth
+            if depth_adjustment:
+                if merge_depth < depth_adjustment:
+                    depth_adjustment = merge_depth
+                merge_depth -= depth_adjustment
+            yield rev_id, '.'.join(map(str, revno)), merge_depth
+
+
+def calculate_view_revisions(branch, start_revision, end_revision, direction,
+        specific_fileid, generate_merge_revisions, allow_single_merge_revision):
+    """Calculate the revisions to view.
+
+    :return: An iterator of (revision_id, dotted_revno, merge_depth) tuples OR
+             a list of the same tuples.
+    """
+    # This method is no longer called by the main code path.
+    # It is retained for API compatibility and may be deprecated
+    # soon. IGC 20090116
+    start_rev_id, end_rev_id = _get_revision_limits(branch, start_revision,
+        end_revision)
+    view_revisions = list(_calc_view_revisions(branch, start_rev_id, end_rev_id,
+        direction, generate_merge_revisions or specific_fileid,
+        allow_single_merge_revision))
     if specific_fileid:
         view_revisions = _filter_revisions_touching_file_id(branch,
             specific_fileid, view_revisions,
             include_merges=generate_merge_revisions)
-
-    # rebase merge_depth - unless there are no revisions or 
-    # either the first or last revision have merge_depth = 0.
+    return _rebase_merge_depth(view_revisions)
+
+
+def _rebase_merge_depth(view_revisions):
+    """Adjust depths upwards so the top level is 0."""
+    # If either the first or last revision have a merge_depth of 0, we're done
     if view_revisions and view_revisions[0][2] and view_revisions[-1][2]:
         min_depth = min([d for r,n,d in view_revisions])
         if min_depth != 0:
@@ -311,21 +536,17 @@
     return view_revisions
 
 
-def _linear_view_revisions(branch):
-    start_revno, start_revision_id = branch.last_revision_info()
-    repo = branch.repository
-    revision_ids = repo.iter_reverse_revision_history(start_revision_id)
-    for num, revision_id in enumerate(revision_ids):
-        yield revision_id, str(start_revno - num), 0
-
-
-def make_log_rev_iterator(branch, view_revisions, generate_delta, search):
+def make_log_rev_iterator(branch, view_revisions, generate_delta, search,
+        file_ids=None, direction='reverse'):
     """Create a revision iterator for log.
 
     :param branch: The branch being logged.
     :param view_revisions: The revisions being viewed.
     :param generate_delta: Whether to generate a delta for each revision.
     :param search: A user text search string.
+    :param file_ids: If non empty, only revisions matching one or more of
+      the file-ids are to be kept.
+    :param direction: the direction in which view_revisions is sorted
     :return: An iterator over lists of ((rev_id, revno, merge_depth), rev,
         delta).
     """
@@ -342,8 +563,14 @@
                 yield (view, None, None)
         log_rev_iterator = iter([_convert()])
     for adapter in log_adapters:
-        log_rev_iterator = adapter(branch, generate_delta, search,
-            log_rev_iterator)
+        # It would be nicer if log adapters were first class objects
+        # with custom parameters. This will do for now. IGC 20090127
+        if adapter == _make_delta_filter:
+            log_rev_iterator = adapter(branch, generate_delta,
+                search, log_rev_iterator, file_ids, direction)
+        else:
+            log_rev_iterator = adapter(branch, generate_delta,
+                search, log_rev_iterator)
     return log_rev_iterator
 
 
@@ -374,7 +601,8 @@
         yield new_revs
 
 
-def _make_delta_filter(branch, generate_delta, search, log_rev_iterator):
+def _make_delta_filter(branch, generate_delta, search, log_rev_iterator,
+    fileids=None, direction='reverse'):
     """Add revision deltas to a log iterator if needed.
 
     :param branch: The branch being logged.
@@ -382,21 +610,86 @@
     :param search: A user text search string.
     :param log_rev_iterator: An input iterator containing all revisions that
         could be displayed, in lists.
+    :param fileids: If non empty, only revisions matching one or more of
+      the file-ids are to be kept.
+    :param direction: the direction in which view_revisions is sorted
     :return: An iterator over lists of ((rev_id, revno, merge_depth), rev,
         delta).
     """
-    if not generate_delta:
+    if not generate_delta and not fileids:
         return log_rev_iterator
-    return _generate_deltas(branch.repository, log_rev_iterator)
-
-
-def _generate_deltas(repository, log_rev_iterator):
-    """Create deltas for each batch of revisions in log_rev_iterator."""
+    return _generate_deltas(branch.repository, log_rev_iterator,
+        generate_delta, fileids, direction)
+
+
+def _generate_deltas(repository, log_rev_iterator, always_delta, fileids,
+    direction):
+    """Create deltas for each batch of revisions in log_rev_iterator.
+    
+    If we're only generating deltas for the sake of filtering against
+    file-ids, we stop generating deltas once all file-ids reach the
+    appropriate life-cycle point. If we're receiving data newest to
+    oldest, then that life-cycle point is 'add', otherwise it's 'remove'.
+    """
+    check_fileids = fileids is not None and len(fileids) > 0
+    if check_fileids:
+        fileid_set = set(fileids)
+        if direction == 'reverse':
+            stop_on = 'add'
+        else:
+            stop_on = 'remove'
+    else:
+        fileid_set = None
     for revs in log_rev_iterator:
+        # If we were matching against fileids and we've run out,
+        # there's nothing left to do
+        if check_fileids and not fileid_set:
+            return
         revisions = [rev[1] for rev in revs]
         deltas = repository.get_deltas_for_revisions(revisions)
-        revs = [(rev[0], rev[1], delta) for rev, delta in izip(revs, deltas)]
-        yield revs
+        new_revs = []
+        for rev, delta in izip(revs, deltas):
+            if check_fileids:
+                if not _delta_matches_fileids(delta, fileid_set, stop_on):
+                    continue
+                elif not always_delta:
+                    # Delta was created just for matching - ditch it
+                    # Note: It would probably be a better UI to return
+                    # a delta filtered by the file-ids, rather than
+                    # None at all. That functional enhancement can
+                    # come later ...
+                    delta = None
+            new_revs.append((rev[0], rev[1], delta))
+        yield new_revs
+
+
+def _delta_matches_fileids(delta, fileids, stop_on='add'):
+    """Check is a delta matches one of more file-ids.
+    
+    :param fileids: a set of fileids to match against.
+    :param stop_on: either 'add' or 'remove' - take file-ids out of the
+      fileids set once their add or remove entry is detected respectively
+    """
+    if not fileids:
+        return False
+    result = False
+    for item in delta.added:
+        if item[1] in fileids:
+            if stop_on == 'add':
+                fileids.remove(item[1])
+            result = True
+    for item in delta.removed:
+        if item[1] in fileids:
+            if stop_on == 'delete':
+                fileids.remove(item[1])
+            result = True
+    if result:
+        return True
+    for l in (delta.modified, delta.renamed, delta.kind_changed):
+        for item in l:
+            if item[1] in fileids:
+                return True
+    return False
 
 
 def _make_revision_objects(branch, generate_delta, search, log_rev_iterator):
@@ -443,6 +736,56 @@
             num = min(int(num * 1.5), 200)
 
 
+def _get_revision_limits(branch, start_revision, end_revision):
+    """Get and check revision limits.
+
+    :param  branch: The branch containing the revisions. 
+
+    :param  start_revision: The first revision to be logged.
+            For backwards compatibility this may be a mainline integer revno,
+            but for merge revision support a RevisionInfo is expected.
+
+    :param  end_revision: The last revision to be logged.
+            For backwards compatibility this may be a mainline integer revno,
+            but for merge revision support a RevisionInfo is expected.
+
+    :return: (start_rev_id, end_rev_id) tuple.
+    """
+    branch_revno, branch_rev_id = branch.last_revision_info()
+    start_rev_id = None
+    if start_revision is None:
+        start_revno = 1
+    else:
+        if isinstance(start_revision, revisionspec.RevisionInfo):
+            start_rev_id = start_revision.rev_id
+            start_revno = start_revision.revno or 1
+        else:
+            branch.check_real_revno(start_revision)
+            start_revno = start_revision
+            start_rev_id = branch.get_rev_id(start_revno)
+
+    end_rev_id = None
+    if end_revision is None:
+        end_revno = branch_revno
+    else:
+        if isinstance(end_revision, revisionspec.RevisionInfo):
+            end_rev_id = end_revision.rev_id
+            end_revno = end_revision.revno or branch_revno
+        else:
+            branch.check_real_revno(end_revision)
+            end_revno = end_revision
+            end_rev_id = branch.get_rev_id(end_revno)
+
+    if branch_revno != 0:
+        if (start_rev_id == _mod_revision.NULL_REVISION
+            or end_rev_id == _mod_revision.NULL_REVISION):
+            raise errors.BzrCommandError('Logging revision 0 is invalid.')
+        if start_revno > end_revno:
+            raise errors.BzrCommandError("Start revision must be older than "
+                                         "the end revision.")
+    return (start_rev_id, end_rev_id)
+
+
 def _get_mainline_revs(branch, start_revision, end_revision):
     """Get the mainline revisions from the branch.
     
@@ -539,6 +882,8 @@
 
     :return: The filtered view_revisions.
     """
+    # This method is no longer called by the main code path.
+    # It may be removed soon. IGC 20090127
     if start_rev_id or end_rev_id:
         revision_ids = [r for r, n, d in view_revisions]
         if start_rev_id:
@@ -655,6 +1000,9 @@
     :return: an iterator of (revision_id, revno, merge_depth)
     (if there is no revno for a revision, None is supplied)
     """
+    # This method is no longer called by the main code path.
+    # It is retained for API compatibility and may be deprecated
+    # soon. IGC 20090127
     if not include_merges:
         revision_ids = mainline_revs[1:]
         if direction == 'reverse':
@@ -731,7 +1079,7 @@
     def __init__(self, rev=None, revno=None, merge_depth=0, delta=None,
                  tags=None, diff=None):
         self.rev = rev
-        self.revno = revno
+        self.revno = str(revno)
         self.merge_depth = merge_depth
         self.delta = delta
         self.tags = tags
@@ -752,10 +1100,10 @@
     - supports_delta must be True if this log formatter supports delta.
         Otherwise the delta attribute may not be populated.  The 'delta_format'
         attribute describes whether the 'short_status' format (1) or the long
-        one (2) sould be used.
+        one (2) should be used.
  
     - supports_merge_revisions must be True if this log formatter supports 
-        merge revisions.  If not, and if supports_single_merge_revisions is
+        merge revisions.  If not, and if supports_single_merge_revision is
         also not True, then only mainline revisions will be passed to the 
         formatter.
 

=== modified file 'bzrlib/tests/blackbox/test_log.py'
--- bzrlib/tests/blackbox/test_log.py	2009-01-29 07:57:43 +0000
+++ bzrlib/tests/blackbox/test_log.py	2009-01-31 04:17:43 +0000
@@ -22,7 +22,7 @@
 
 from bzrlib import osutils
 from bzrlib.tests.blackbox import ExternalBase
-from bzrlib.tests import TestCaseInTempDir, TestCaseWithTransport
+from bzrlib.tests import KnownFailure, TestCaseInTempDir, TestCaseWithTransport
 from bzrlib.tests.test_log import (
     normalize_log,
     )
@@ -482,6 +482,10 @@
 """)
 
     def test_merges_nonsupporting_formatter(self):
+        # This "feature" of log formatters is madness. If a log
+        # formatter cannot display a dotted-revno, it ought to ignore it.
+        # Otherwise, a linear sequence is always expected to be handled now.
+        raise KnownFailure('log formatters must support linear sequences now')
         self._prepare()
         err_msg = 'Selected log formatter only supports mainline revisions.'
         # The single revision case is tested in the core tests

=== modified file 'bzrlib/tests/test_log.py'
--- bzrlib/tests/test_log.py	2009-01-29 06:18:25 +0000
+++ bzrlib/tests/test_log.py	2009-01-31 04:17:43 +0000
@@ -1314,11 +1314,8 @@
         rev_4b = rev_from_rev_id('4b', wt.branch)
         self.assertEqual([('3c', '3', 0), ('3a', '2.1.1', 1)],
                           view_revs(rev_3a, rev_4b, 'f-id', 'reverse'))
-        # Note that the depth is 0 for 3a because depths are normalized, but
-        # there is still a bug somewhere... most probably in
-        # _filter_revision_range and/or get_view_revisions still around a bad
-        # use of reverse_by_depth
-        self.assertEqual([('3a', '2.1.1', 0)],
+        # Note: 3c still appears before 3a here because of depth-based sorting
+        self.assertEqual([('3c', '3', 0), ('3a', '2.1.1', 1)],
                           view_revs(rev_3a, rev_4b, 'f-id', 'forward'))
 
 

